@@ -53,18 +53,17 @@ module Language.C.Analysis.TravMonad (
 )
 where
 import Language.C.Data
-import Language.C.Data.Ident
 import Language.C.Data.RList as RList
-import Language.C.Syntax
 
 import Language.C.Analysis.Builtins
 import Language.C.Analysis.SemError
 import Language.C.Analysis.SemRep
+import Language.C.Analysis.TypeUtils (sameType)
 import Language.C.Analysis.DefTable hiding (enterBlockScope,leaveBlockScope,
                                             enterFunctionScope,leaveFunctionScope)
 import qualified Language.C.Analysis.DefTable as ST
 
-import Data.IntMap (insert, lookup)
+import Data.IntMap (insert)
 import Data.Maybe
 import Control.Applicative (Applicative(..))
 import Control.Monad (liftM, ap)
@@ -119,7 +118,7 @@ checkRedef subject new_decl redecl_status =
 handleTagDecl :: (MonadCError m, MonadSymtab m) => TagFwdDecl -> m ()
 handleTagDecl decl = do
     redecl <- withDefTable $ declareTag (sueRef decl) decl
-    checkRedef (show $ sueRef decl) decl redecl
+    checkRedef (sueRefToString $ sueRef decl) decl redecl
 
 -- | define the given composite type or enumeration
 -- If there is a declaration visible, overwrite it with the definition.
@@ -128,20 +127,27 @@ handleTagDecl decl = do
 handleTagDef :: (MonadTrav m) => TagDef -> m ()
 handleTagDef def = do
     redecl <- withDefTable $ defineTag (sueRef def) def
-    checkRedef (show $ sueRef def) def redecl
+    checkRedef (sueRefToString $ sueRef def) def redecl
     handleDecl (TagEvent def)
 
 handleEnumeratorDef :: (MonadCError m, MonadSymtab m) => Enumerator ->  m ()
 handleEnumeratorDef enumerator = do
     let ident = declIdent enumerator
     redecl <- withDefTable $ defineScopedIdent ident (EnumeratorDef enumerator)
-    checkRedef (show ident) ident redecl
+    checkRedef (identToString ident) ident redecl
     return ()
 
 handleTypeDef :: (MonadTrav m) => TypeDef -> m ()
-handleTypeDef typeDef@(TypeDef ident _ _ _) = do
+handleTypeDef typeDef@(TypeDef ident t1 _ _) = do
     redecl <- withDefTable $ defineTypeDef ident typeDef
-    checkRedef (show ident) typeDef redecl
+    -- C11 6.7/3 If an identifier has no linkage, there shall be no more than
+    -- one declaration of the identifier (in a declarator or type specifier)
+    -- with the same scope and in the same name space, except that: a typedef
+    -- name may be redefined to denote the same type as it currently does,
+    -- provided that type is not a variably modified type;
+    case redecl of
+      Redeclared (Left (TypeDef _ t2 _ _)) | sameType t1 t2 -> return ()
+      _ -> checkRedef (identToString ident) typeDef redecl
     handleDecl (TypeDefEvent typeDef)
     return ()
 
@@ -151,16 +157,16 @@ handleAsmBlock asm = handleDecl (AsmEvent asm)
 redefErr :: (MonadCError m, CNode old, CNode new) =>
             Ident -> ErrorLevel -> new -> old -> RedefKind -> m ()
 redefErr name lvl new old kind =
-  throwTravError $ redefinition lvl (show name) kind (nodeInfo new) (nodeInfo old)
+  throwTravError $ redefinition lvl (identToString name) kind (nodeInfo new) (nodeInfo old)
 
 -- TODO: unused
-checkIdentTyRedef :: (MonadCError m) => IdentEntry -> (DeclarationStatus IdentEntry) -> m ()
-checkIdentTyRedef (Right decl) status = checkVarRedef decl status
-checkIdentTyRedef (Left tydef) (KindMismatch old_def) =
+_checkIdentTyRedef :: (MonadCError m) => IdentEntry -> (DeclarationStatus IdentEntry) -> m ()
+_checkIdentTyRedef (Right decl) status = checkVarRedef decl status
+_checkIdentTyRedef (Left tydef) (KindMismatch old_def) =
   redefErr (identOfTypeDef tydef) LevelError tydef old_def DiffKindRedecl
-checkIdentTyRedef (Left tydef) (Redeclared old_def) =
+_checkIdentTyRedef (Left tydef) (Redeclared old_def) =
   redefErr (identOfTypeDef tydef) LevelError tydef old_def DuplicateDef
-checkIdentTyRedef (Left _tydef) _ = return ()
+_checkIdentTyRedef (Left _tydef) _ = return ()
 
 -- Check whether it is ok to declare a variable already in scope
 checkVarRedef :: (MonadCError m) => IdentDecl -> (DeclarationStatus IdentEntry) -> m ()
@@ -185,19 +191,19 @@ checkVarRedef def redecl =
         _ -> return ()
     where
     redefVarErr old_def kind = redefErr (declIdent def) LevelError def old_def kind
-    linkageErr def old_def =
-        case (declLinkage def, declLinkage old_def) of
-            (NoLinkage, _) -> redefErr (declIdent def) LevelError  def old_def NoLinkageOld
-            otherwise      -> redefErr (declIdent def) LevelError  def old_def DisagreeLinkage
+    linkageErr new_def old_def =
+        case (declLinkage new_def, declLinkage old_def) of
+            (NoLinkage, _) -> redefErr (declIdent new_def) LevelError new_def old_def NoLinkageOld
+            _              -> redefErr (declIdent new_def) LevelError new_def old_def DisagreeLinkage
 
     new_ty = declType def
     canBeOverwritten (Declaration _) = True
     canBeOverwritten (ObjectDef od)  = isTentative od
     canBeOverwritten _               = False
-    agreeOnLinkage def old_def
+    agreeOnLinkage new_def old_def
         | declStorage old_def == FunLinkage InternalLinkage = True
-        | not (hasLinkage $ declStorage def) || not (hasLinkage $ declStorage old_def) = False
-        | (declLinkage def) /= (declLinkage old_def) = False
+        | not (hasLinkage $ declStorage new_def) || not (hasLinkage $ declStorage old_def) = False
+        | (declLinkage new_def) /= (declLinkage old_def) = False
         | otherwise = True
 
 -- | handle variable declarations (external object declarations and function prototypes)
@@ -251,7 +257,7 @@ handleObjectDef :: (MonadTrav m) => Bool -> Ident -> ObjDef -> m ()
 handleObjectDef local ident obj_def = do
     let def = ObjectDef obj_def
     redecl <- withDefTable $
-              defineScopedIdentWhen (\old -> shouldOverride def old) ident def
+              defineScopedIdentWhen (shouldOverride def) ident def
     checkVarRedef def redecl
     handleDecl ((if local then LocalEvent else DeclEvent) def)
     where
@@ -356,7 +362,7 @@ handleTravError a = liftM Just a `catchTravError` (\e -> recordError e >> return
 
 -- | check wheter non-recoverable errors occurred
 hadHardErrors :: [CError] -> Bool
-hadHardErrors = (not . null . filter isHardError)
+hadHardErrors = any isHardError
 
 -- | raise an error caused by a malformed AST
 astError :: (MonadCError m) => NodeInfo -> String -> m a
